<?php

namespace com\fwdekker\deathnotifier;

use Error;
use InvalidArgumentException;


/**
 * The application's configuration, read lazily from configuration files.
 *
 * Contains global state, but that's fine since it's read-only (except for some internal functions for testing).
 */
class Config
{
    /**
     * @var array<string, mixed>|null the application's configuration, or `null` if it has not been read yet
     */
    private static ?array $config = null;


    /**
     * Returns the property at {@see $index}.
     *
     * To get the property `name` from section `section`, set `$index` to `"name.section"`. To return the entire section
     * `section` as an array, set `$index` to `"section"`.
     *
     * @param string|null $index the index of the property to return, using `.` as a nesting delimiter; or `null` to
     * return the entire configuration
     * @return mixed|null the property at {@see $index}, or `null` if there is no such property
     */
    public static function get(?string $index = null): mixed
    {
        if (self::$config === null)
            self::$config = self::read_config_from_file();

        if ($index === null)
            return self::$config;

        $output = self::$config;
        foreach (explode(".", $index) as $key) {
            if (!isset($output[$key]))
                return null;

            $output = $output[$key];
        }
        return $output;
    }

    /**
     * Returns `true` if and only if the property at {@see $index} is not `null`.
     *
     * @param string $index the index of the property to check, using `.` as a nesting delimiter
     * @return bool `true` if and only if the property at {@see $index} is not `null`
     */
    public static function has(string $index): bool
    {
        return self::get($index) !== null;
    }


    /**
     * Sets the property at {@see $index} to {@see $value}.
     *
     * This function is intended for testability. It should never be used in production code.
     *
     * @param string $index the index to set to {@see $value}, using `.` as a nesting delimiter
     * @param mixed $value the value to write at {@see $index}
     * @return void
     * @see Config::_reset()
     */
    public static function _set(string $index, mixed $value): void
    {
        self::get();

        $keys = explode(".", $index);
        $last_key = array_pop($keys);

        $output = &self::$config;
        foreach ($keys as $key) {
            if (!isset($output[$key]))
                $output[$key] = [];

            $output = &$output[$key];
        }
        $output[$last_key] = $value;
    }

    /**
     * Resets all manual modifications to the application's configuration.
     *
     * This function is intended for testability. It should never be used in production code.
     *
     * @return void
     * @see Config::_set()
     */
    public static function _reset(): void
    {
        self::$config = null;
    }


    /**
     * Reads the configuration file and overrides it with the user's custom values.
     *
     * @return array<string, mixed> the application's configuration
     */
    private static function read_config_from_file(): array
    {
        $config = parse_ini_file("config.default.ini.php", process_sections: true, scanner_mode: INI_SCANNER_TYPED);
        if ($config === false) throw new InvalidArgumentException("Invalid `config.default.ini.php` file.");

        if (file_exists("config.ini.php")) {
            $config_custom = parse_ini_file("config.ini.php", process_sections: true, scanner_mode: INI_SCANNER_TYPED);
            if ($config_custom === false) throw new InvalidArgumentException("Invalid `config.ini.php` file.");

            $config = Util::array_merge_recursive_distinct($config, $config_custom);

            // Check file permissions
            if (!$config["security"]["allow_config_insecure_permissions"]) {
                $perms = fileperms("config.ini.php");
                if ($perms === false)
                    throw new Error("Failed to read file permissions for `config.ini.php` even though the file exists.");

                $perms = substr(decoct($perms), 3);
                if ($perms !== "600")
                    throw new InvalidArgumentException(
                        "Insecure file permissions for `config.ini.php`. " .
                        "Should be 600 but was $perms."
                    );
            }
        }

        return $config;
    }
}
