<?php

namespace com\fwdekker\deathnotifier;

use com\fwdekker\deathnotifier\user\UserList;
use Exception;


/**
 * Starts a new user session, or continues an existing one.
 */
class StartSessionAction extends Action
{
    /**
     * @var UserList the list to validate the session against
     */
    private readonly UserList $user_list;


    /**
     * Constructs a new `StartSessionAction`.
     *
     * @param UserList $user_list the list to validate the session against
     */
    public function __construct(UserList $user_list)
    {
        $this->user_list = $user_list;
    }


    /**
     * Starts a new user session, or continues an existing one.
     *
     * Does not require a CSRF token. Does not require the user to be logged in or out.
     *
     * @param array<int|string, mixed> $inputs ignored
     * @return array{"logged_in": bool, "global_message"?: string} whether the user is logged in, and the message to be
     * displayed at the top of the page, if any
     * @throws UnexpectedException if no new CSRF token could be generated
     */
    public function handle(array $inputs): array
    {
        $payload = [];

        // Check if logged in
        if (!isset($_SESSION["uuid"])) {
            $payload["logged_in"] = false;
        } else if ($this->user_list->has_user_with_uuid($_SESSION["uuid"])) {
            $payload["logged_in"] = true;
        } else {
            // User already logged in, but user was deleted
            session_destroy();
            session_start();
            try {
                $_SESSION["token"] = Util::generate_csrf_token();
            } catch (Exception $exception) {
                throw new UnexpectedException(
                    "Failed to generate new CSRF token. Please try again later.",
                    previous: $exception
                );
            }

            $payload["logged_in"] = false;
        }

        // Read global message
        if (Config::has("server.global_message") && trim(Config::get("server.global_message")) !== "")
            $payload["global_message"] = trim(Config::get("server.global_message"));

        return $payload;
    }
}
