<?php

namespace com\fwdekker\deathnotifier;

use com\fwdekker\deathnotifier\validation\InvalidValueException;
use Exception;
use JsonException;


/**
 * Helper functions.
 */
class Util
{
    /**
     * Parses POST values from JSON-based inputs.
     *
     * @return array<int|string, mixed> the parsed POSTed values, or an empty array if no values were POSTed
     * @throws InvalidValueException if there are no POST input values, or if the POST input is not valid JSON
     */
    static function parse_post(): array
    {
        $post_input = file_get_contents("php://input");
        if ($post_input === false || trim($post_input) === "")
            return [];

        try {
            $post = json_decode($post_input, associative: true, flags: JSON_THROW_ON_ERROR);
            if ($post === null)
                throw new InvalidValueException("Malformed request: POST data is `null`.");

            return $post;
        } catch (JsonException) {
            throw new InvalidValueException("Malformed request: POST data could not be parsed as JSON.");
        }
    }

    /**
     * Parses `$argv` into an array, similar to `$_GET`.
     *
     * Code from https://www.php.net/manual/en/features.commandline.php#108883.
     *
     * @return array<int|string, mixed> the parsed CLI inputs, or an empty array if no values were passed on the CLI
     */
    static function parse_cli(): array
    {
        if (!isset($_SERVER["argc"]) || $_SERVER["argc"] <= 1) return [];

        parse_str(implode("&", array_slice($_SERVER["argv"], 1)), $output);
        return $output;
    }


    /**
     * Generates an appropriate CSRF token.
     *
     * @return string the generated CSRF token
     * @throws Exception if the CSRF token could not be generated
     */
    static function generate_csrf_token(): string
    {
        return bin2hex(random_bytes(32));
    }


    /**
     * Returns the number of minutes until {@see $timestamp} was {@see $interval} minutes ago.
     *
     * For example, if {@see $timestamp} was 5 minutes ago, and {@see $interval} is 7, then this function returns 2.
     *
     * @param int $timestamp the timestamp at which some event occurred
     * @param int $interval the number of minutes to measure against
     * @return int the number of minutes until {@see $timestamp} was {@see $interval} minutes ago
     */
    static function minutes_until_interval_elapsed(int $timestamp, int $interval): int
    {
        return $interval - ((time() - $timestamp) / 60);
    }

    /**
     * Recursively merges arrays, while overwriting other types.
     *
     * Functions similar to `array_merge_recursive`, except that if two values are encountered at least one of which is not
     * an array, the value of `array2` is taken, instead of taking an array of both values.
     *
     * If a key exists in `array1` but not in `array2`, then the value of `array1` is used. If a key exists in `array2` but
     * not in `array1`, then the value of `array2` is used. If a key exists in both `array1` and `array2`, and both values
     * are arrays, this function is applied recursively, effectively using a merged array containing the values of both
     * arrays' arrays. If a key exists in both `array1` and `array2`, and at least one of the values is not an array, the
     * value of `array2` is used.
     *
     * Taken from `https://www.php.net/manual/en/function.array-merge-recursive.php#92195`.
     *
     * @param mixed[] $array1 the base array to merge into
     * @param mixed[] $array2 the array to merge into `array1`
     * @return mixed[] the recursively merged array
     * @author Daniel <daniel (at) danielsmedegaardbuus (dot) dk>
     * @author Gabriel Sobrinho <gabriel (dot) sobrinho (at) gmail (dot) com>
     */
    static function array_merge_recursive_distinct(array $array1, array $array2): array
    {
        $merged = $array1;

        foreach ($array2 as $key => $value)
            if (is_array($value) && isset ($merged[$key]) && is_array($merged[$key]))
                $merged[$key] = Util::array_merge_recursive_distinct($merged[$key], $value);
            else
                $merged[$key] = $value;

        return $merged;
    }

}
