<?php

namespace com\fwdekker\deathnotifier\user;

use com\fwdekker\deathnotifier\Action;
use com\fwdekker\deathnotifier\UnexpectedException;
use com\fwdekker\deathnotifier\validation\InvalidTypeException;
use com\fwdekker\deathnotifier\validation\InvalidValueException;
use com\fwdekker\deathnotifier\validation\IsValidCsrfTokenRule;
use com\fwdekker\deathnotifier\validation\RuleSet;
use com\fwdekker\deathnotifier\validation\LoginValidator;


/**
 * Returns the user's public data.
 */
class GetPublicUserDataAction extends Action
{
    /**
     * @var UserList the list to retrieve public data from
     */
    private readonly UserList $user_list;


    /**
     * Constructs a new `GetPublicUserDataAction`.
     *
     * @param UserList $user_list the list to retrieve public data from
     */
    public function __construct(UserList $user_list)
    {
        $this->user_list = $user_list;
    }


    /**
     * Returns the user's public data.
     *
     * Requires that the user is logged in and that a valid CSRF token is present.
     *
     * @param array<int|string, mixed> $inputs `"token": string`: a valid CSRF token
     * @return array{"email": string, "email_verified": bool, "email_notifications_enabled": bool,
     *     "password_last_change": int} the user's public data
     * @throws InvalidTypeException if any of the inputs has the incorrect type
     * @throws InvalidValueException if the user is not logged in or if no valid CSRF token is present
     * @throws UnexpectedException if the current user has been deleted
     */
    public function handle(array $inputs): array
    {
        (new LoginValidator(validate_logged_in: true))->check($_SESSION);
        (new RuleSet(["token" => [new IsValidCsrfTokenRule()]]))->check($inputs);

        $user_data = $this->user_list->get_user_by_uuid($_SESSION["uuid"]);
        if ($user_data === null)
            throw new UnexpectedException("Failed to retrieve account data. Please refresh the page and try again.");

        return [
            "email" => $user_data["email"],
            "email_verified" => $user_data["email_verification_token"] === null,
            "email_notifications_enabled" => $user_data["email_notifications_enabled"] === 1,
            "password_last_change" => $user_data["password_last_change"],
        ];
    }
}
