<?php

namespace com\fwdekker\deathnotifier\validation;

use InvalidArgumentException;


/**
 * Validates that the input is of the specific length.
 */
class HasStringLengthRule extends Rule
{
    /**
     * @var int|null the minimum length (inclusive), or `null` if there is no minimum length
     */
    private readonly ?int $min_length;
    /**
     * @var int|null the maximum length (inclusive), or `null` if there is no maximum length
     */
    private readonly ?int $max_length;


    /**
     * Constructs a new `HasLengthRule`.
     *
     * @param int|null $min_length the minimum length (inclusive), or `null` if there is no minimum length
     * @param int|null $max_length the maximum length (inclusive), or `null` if there is no maximum length
     */
    public function __construct(?int $min_length = null, ?int $max_length = null)
    {
        if ($min_length !== null && $max_length !== null && $min_length > $max_length)
            throw new InvalidArgumentException("Minimum length should not exceed maximum length.");
        if ($min_length === null && $max_length === null)
            throw new InvalidArgumentException("Either minimum length or maximum length should be non-null.");

        $this->min_length = $min_length;
        $this->max_length = $max_length;
    }


    /**
     * Validates that the input is of the specific length.
     *
     * @param array<int|string, mixed> $inputs the list of inputs in which the value at {@see $key} should be checked
     * @param string $key the key in {@see $inputs} of the input to check
     * @return void if the checked input is of the specified length
     * @throws InvalidTypeException if the checked input is not set or is not a string
     * @throws InvalidValueException if the checked input is not of the specified length
     */
    public function check(array $inputs, string $key): void
    {
        if (!isset($inputs[$key]))
            throw new InvalidTypeException("Required input '$key' not set.");
        if (!is_string($inputs[$key]))
            throw new InvalidTypeException("Input '$key' should be string, but is " . gettype($inputs[$key]) . ".");

        if ($this->min_length !== null && strlen($inputs[$key]) < $this->min_length) {
            $characters = $this->min_length !== 1 ? "characters" : "character";
            throw new InvalidValueException("Use at least $this->min_length $characters.", $key);
        }
        if ($this->max_length !== null && strlen($inputs[$key]) > $this->max_length) {
            $characters = $this->min_length !== 1 ? "characters" : "character";
            throw new InvalidValueException("Use at most $this->max_length $characters.", $key);
        }
    }
}
