<?php

namespace com\fwdekker\deathnotifier\wikipedia;

use ArrayIterator;
use IteratorAggregate;
use Traversable;


/**
 * Collects redirects and normalizations of Wikipedia article titles.
 *
 * Can contain reflexive redirects, i.e. a "redirect" from a title to the exact same title.
 *
 * @implements IteratorAggregate<string, string>
 */
class Redirects implements IteratorAggregate
{
    /**
     * @var array<string, string> the known redirects
     */
    private array $redirects = [];


    /**
     * Adds new redirects and resolves double redirects, and returns a new `Redirects` containing the old and new
     * redirects.
     *
     * If there already exists a redirect `a -> b`, and the redirect `b -> c` is added, then `a -> b` is changed to
     * `a -> c`.
     *
     * If there already exists a redirect `b -> c`, and the redirect `a -> b` is added, then `a -> b` is added as
     * `a -> c`.
     *
     * @param array<string, string> $redirects the redirects to add
     * @return Redirects a new `Redirects` containing the old and new redirects
     */
    public function add(array $redirects): Redirects
    {
        $all_redirects = $this->redirects;

        foreach ($redirects as $from => $to) {
            $to_redirects_to = $all_redirects[$to] ?? $to;
            $all_redirects[$from] = $to_redirects_to;

            $redirects_to_from = array_keys($all_redirects, $from);
            foreach ($redirects_to_from as $redirect_to_from)
                $all_redirects[$redirect_to_from] = $to_redirects_to;
        }

        $new_object = new Redirects();
        $new_object->redirects = $all_redirects;
        return $new_object;
    }

    /**
     * Returns the page that {@see from} is redirect to, or {@see from} if there is no redirect.
     *
     * @param string $from the page to resolve the redirect of
     * @return string the page that {@see from} is redirect to, or {@see from} if there is no redirect
     */
    public function resolve(string $from): string
    {
        return $this->redirects[$from] ?? $from;
    }


    /**
     * Returns `true` if and only if there is at least one redirect that is not reflexive.
     *
     * @return bool `true` if and only if there is at least one redirect that is not reflexive
     */
    public function has_non_reflexive_redirects(): bool
    {
        return array_keys($this->redirects) !== array_values($this->redirects);
    }


    /**
     * Returns a {@see Traversable} to iterate over the redirects.
     *
     * @return Traversable<string, string> a {@see Traversable} to iterate over the redirects
     */
    public function getIterator(): Traversable
    {
        return new ArrayIterator($this->redirects);
    }
}
